package org.openmuc.jdlms;

import java.io.IOException;

import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.connectionSettings.TcpServerSettings;
import org.openmuc.jdlms.internal.transportlayer.ServerTcpLayer;
import org.openmuc.jdlms.internal.transportlayer.ServerTransportLayer;

/**
 * Class representing a physical device (DLMS/COSEM server).
 */
public class DlmsServer implements AutoCloseable {

    private final ServerTransportLayer serverTransportLayer;

    private DlmsServer(ServerTransportLayer serverTransportLayer) {
        this.serverTransportLayer = serverTransportLayer;
    }

    /**
     * Starts the DLMS/COSEM server.
     * 
     * @throws IOException
     *             if an exception occurs during the start of the server.
     */
    public void start() throws IOException {
        serverTransportLayer.start();
    }

    /**
     * Stops the server immediately. Releases all acquired resources.
     */
    @Override
    public void close() throws IOException {
        this.serverTransportLayer.close();
    }

    /**
     * Sends disconnect messages to all connected clients.
     * 
     * @throws IOException
     *             if an exception occurs, while releasing the resources.
     */
    public void shutdown() throws IOException {
        // TODO:
        close();
    }

    /**
     * Create a new TCP server builder.
     * 
     * @param port
     *            the TCP port the server starts listening on.
     * @return a new TcpServerBuilder.
     */
    public static TcpServerBuilder tcpServerBuilder(int port) {
        return new TcpServerBuilder(port);
    }

    public static HdlcSerialServerBuilder hdlcSerialServerBuilder(String serialPort) {
        return new HdlcSerialServerBuilder(serialPort);
    }

    public static class HdlcSerialServerBuilder extends ServerBuilder<HdlcSerialServerBuilder> {

        private HdlcSerialServerBuilder(String serialPort) {
            throw new UnsupportedOperationException("HdlcSerial not yet implemented.");
        }

        @Override
        public DlmsServer build() {
            return null;
        }

    }

    /**
     * Builder to create a TCP physical device/server.
     */
    public static class TcpServerBuilder extends ServerBuilder<TcpServerBuilder> {

        private int port;

        private TcpServerBuilder(int port) {
            this.port = port;
        }

        /**
         * The port a client may access the server.
         * 
         * @param port
         *            the TCP port the server starts listening on.
         * @return the current builder instance.
         */
        public TcpServerBuilder setTcpPort(int port) {
            this.port = port;
            return this;
        }

        @Override
        public DlmsServer build() {
            DataDirectory dataDirectory = parseLogicalDevices();

            TcpServerSettings settings = new TcpServerSettings(this.port);
            setPropertiesTo(settings);

            return new DlmsServer(new ServerTcpLayer(dataDirectory, settings));
        }

    }

}
