package org.openmuc.jdlms.internal.lnassociation;

import java.io.EOFException;
import java.io.IOException;
import java.net.SocketTimeoutException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Semaphore;

import org.openmuc.jdlms.Authentication;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.ConnectionData;
import org.openmuc.jdlms.internal.DataDirectory;
import org.openmuc.jdlms.internal.DataDirectory.DlmsLogicalDevice;
import org.openmuc.jdlms.internal.ServiceError;
import org.openmuc.jdlms.internal.StateError;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu.Choices;
import org.openmuc.jdlms.internal.sessionlayer.ServerSessionLayer;

public class Association implements Runnable {

    private final DataDirectory directory;
    private final ServerSessionLayer sessionLayer;

    private final Semaphore semaphore;
    private final Long connectionId;
    private final ConnectionData connectionData;

    public Association(DataDirectory directory, ServerSessionLayer sessionLayer, Semaphore semaphore,
            Long connectionId) {
        this.directory = directory;
        this.sessionLayer = sessionLayer;
        this.semaphore = semaphore;
        this.connectionId = connectionId;

        this.connectionData = new ConnectionData();
        this.directory.addConnection(connectionId, this.connectionData);

    }

    @Override
    public void run() {
        AssociationMessenger associationMessenger = new AssociationMessenger(connectionData, directory, sessionLayer,
                connectionId);

        try {
            byte[] payload = this.sessionLayer.readNextMessage();

            RequestProcessorData requestProcessorData = new RequestProcessorData(this.sessionLayer.getLogicalDeviceId(),
                    connectionId, directory, connectionData);

            this.connectionData.clientId = this.sessionLayer.getClientId();
            DlmsLogicalDevice dlmsLogicalDevice = this.directory
                    .getLogicalDeviceFor(this.sessionLayer.getLogicalDeviceId());

            Map<Integer, Authentication> restrictions = dlmsLogicalDevice.getLogicalDevice().getRestrictions();
            this.connectionData.authentication = restrictions.get(sessionLayer.getClientId());

            APdu aarqAPdu = new InitialmessageProcessor(connectionData, dlmsLogicalDevice)
                    .processInitialMessage(payload);
            associationMessenger.encodeAndSend(aarqAPdu);

            Map<COSEMpdu.Choices, RequestProcessor> requestProcessors = buildRequestProcessors(associationMessenger,
                    requestProcessorData);

            while (true) {
                readNextMessage(associationMessenger, requestProcessors);
            }

        } catch (GenericAssociationException e) {
            e.printStackTrace();
            try {
                associationMessenger.encodeAndSend(e.getErrorMessageApdu());
            } catch (IOException e1) {
                e1.printStackTrace();
                // ignore any exception here..
            }
        } catch (IOException e) {
            if (e instanceof EOFException) {
                System.out.println("client closed the connection..");
                return;
            }
            if (e instanceof SocketTimeoutException) {
                System.out.println("client was too slow.");
                return;
            }

            e.printStackTrace();
            // TODO ignore??
        } finally {
            this.semaphore.release();
            this.directory.removeConnection(this.connectionId);
            try {
                sessionLayer.close();
            } catch (IOException e1) {
                // ignore
            }
            System.out.println("closed");
        }
    }

    private Map<COSEMpdu.Choices, RequestProcessor> buildRequestProcessors(AssociationMessenger associationMessenger,
            RequestProcessorData requestProcessorData) {
        Map<COSEMpdu.Choices, RequestProcessor> requestProcessors = new HashMap<>();
        requestProcessors.put(COSEMpdu.Choices.ACTION_REQUEST,
                new ActionRequestProcessor(associationMessenger, requestProcessorData));
        requestProcessors.put(COSEMpdu.Choices.GET_REQUEST,
                new GetRequestProcessor(associationMessenger, requestProcessorData));
        requestProcessors.put(COSEMpdu.Choices.SET_REQUEST,
                new SetRequestProcessor(associationMessenger, requestProcessorData));
        return requestProcessors;
    }

    private void readNextMessage(AssociationMessenger associationMessenger,
            Map<Choices, RequestProcessor> requestProcessors) throws IOException, AssociationException {

        COSEMpdu cosemPdu = associationMessenger.readNextCosemPdu();

        if (!this.connectionData.authenticated && cosemPdu.getChoiceIndex() != COSEMpdu.Choices.ACTION_REQUEST) {
            throw new AssociationException(StateError.SERVICE_NOT_ALLOWED, ServiceError.OPERATION_NOT_POSSIBLE);
        }

        RequestProcessor requestProcessor = requestProcessors.get(cosemPdu.getChoiceIndex());
        if (requestProcessor != null) {
            requestProcessor.processRequest(cosemPdu);
        }
        else {
            // TODO handle other requests..
        }

        if (!this.connectionData.authenticated) {
            throw new IOException("Client failed to authenticate..");
        }

    }

}
