package org.openmuc.jdlms;

import org.openmuc.jdlms.internal.AuthenticationMechanism;

/**
 * Authentication object used to restrict a connection on a logical device.
 * 
 * @see LogicalDevice
 */
public class Authentication {

    byte[] authenticationKey;
    private final AuthenticationMechanism authenticationMechanism;
    byte[] globalEncryptionKey;
    private final CryptographicAlgorithm cryptographicAlgorithm;

    private Authentication(AuthenticationMechanism authenticationMechanism, byte[] authenticationKey,
            byte[] globalEncryptionKey, CryptographicAlgorithm cryptographicAlgorithm) {
        this.authenticationMechanism = authenticationMechanism;
        this.authenticationKey = authenticationKey;
        this.globalEncryptionKey = globalEncryptionKey;
        this.cryptographicAlgorithm = cryptographicAlgorithm;
    }

    /**
     * Creates a Low Level Security (LLS) that uses a pre-shared password for authentication.
     * 
     * @param authenticationKey
     *            the password.
     * @return a new Authentication object.
     */
    public static Authentication newPasswordAuthentication(byte[] authenticationKey) {
        return new Authentication(AuthenticationMechanism.LOW, authenticationKey, null, null);
    }

    /**
     * Creates a Low Level Security (LLS) that uses a pre-shared password for authentication and using encryption.
     * 
     * @param authenticationKey
     *            the password.
     * @param globalEncryptionKey
     *            the symmetrical encryption key.
     * @param cryptographicAlgorithm
     *            the algorithm used to encrypt the DLMS messages.
     * @return a new Authentication object.
     */
    public static Authentication newPasswordAuthentication(byte[] authenticationKey, byte[] globalEncryptionKey,
            CryptographicAlgorithm cryptographicAlgorithm) {
        return new Authentication(AuthenticationMechanism.LOW, authenticationKey, globalEncryptionKey,
                cryptographicAlgorithm);
    }

    /**
     * Creates a Authentication without any restrictions.
     * 
     * @return a new Authentication object.
     */
    public static Authentication newNoneAuthentication() {
        return new Authentication(AuthenticationMechanism.NONE, null, null, null);
    }

    /**
     * Creates a Authentication object without authentication and using encryption.
     * 
     * @param authenticationKey
     *            authentication key for encryption.
     * @param globalEncryptionKey
     *            the symmetrical encryption key.
     * @param cryptographicAlgorithm
     *            the algorithm used to encrypt the DLMS messages.
     * @return a new Authentication object.
     */
    public static Authentication newNoneAuthentication(byte[] authenticationKey, byte[] globalEncryptionKey,
            CryptographicAlgorithm cryptographicAlgorithm) {
        return new Authentication(AuthenticationMechanism.NONE, authenticationKey, globalEncryptionKey,
                cryptographicAlgorithm);
    }

    /**
     * Use High Level Security (HLS) 3 that uses MD5.
     * 
     * @param authenticationKey
     *            the password
     * @return a new Authentication object.
     */
    public static Authentication newMd5Authentication(byte[] authenticationKey) {
        return new Authentication(AuthenticationMechanism.HLS3_MD5, authenticationKey, null, null);
    }

    /**
     * Use High Level Security (HLS) 3 that uses MD5 and using encryption.
     * 
     * @param authenticationKey
     *            the password.
     * @param globalEncryptionKey
     *            the symmetrical encryption key.
     * @param cryptographicAlgorithm
     *            the algorithm used to encrypt the DLMS messages.
     * @return a new Authentication object.
     */
    public static Authentication newMd5Authentication(byte[] authenticationKey, byte[] globalEncryptionKey,
            CryptographicAlgorithm cryptographicAlgorithm) {
        return new Authentication(AuthenticationMechanism.HLS3_MD5, authenticationKey, globalEncryptionKey,
                cryptographicAlgorithm);
    }

    /**
     * Create a new GMAC Authentication.
     * 
     * @param authenticationKey
     *            the key used to authenticate.
     * @param encryptionKey
     *            the symmetrical encryption key.
     * @return a new Authentication object.
     */
    public static Authentication newGmacAuthentication(byte[] authenticationKey, byte[] encryptionKey) {
        return new Authentication(AuthenticationMechanism.HLS5_GMAC, authenticationKey, encryptionKey, null);
    }

    /**
     * Create a new GMAC Authentication using encryption.
     * 
     * <p>
     * GMAC, an authentication-only variant of
     * <a href="https://en.wikipedia.org/wiki/Galois/Counter_Mode">Galois/Counter_Mode</a> for symmetric key
     * cryptographic block ciphers.
     * </p>
     * 
     * @param authenticationKey
     *            the key used to authenticate.
     * @param encryptionKey
     *            the symmetrical encryption key.
     * @param cryptographicAlgorithm
     *            the algorithm used to encrypt the DLMS messages.
     * @return a new Authentication object.
     */
    public static Authentication newGmacAuthentication(byte[] authenticationKey, byte[] encryptionKey,
            CryptographicAlgorithm cryptographicAlgorithm) {
        return new Authentication(AuthenticationMechanism.HLS5_GMAC, authenticationKey, encryptionKey,
                cryptographicAlgorithm);
    }

    public CryptographicAlgorithm getCryptographicAlgorithm() {
        return cryptographicAlgorithm;
    }

    public byte[] getAuthenticationKey() {
        return authenticationKey;
    }

    public AuthenticationMechanism getAuthenticationMechanism() {
        return authenticationMechanism;
    }

    public byte[] getGlobalEncryptionKey() {
        return globalEncryptionKey;
    }

    /**
     * Cryptographic algorithm identify the algorithm for which a derived secret symmetrical key will be used.
     */
    public enum CryptographicAlgorithm {
        /**
         * Key wrapping using AES-128 key wrap.
         */
        AES_GMC_128(0);

        private int id;

        private CryptographicAlgorithm(int id) {
            this.id = id;
        }

        public int getId() {
            return id;
        }
    }

}
