/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms;

import java.io.IOException;
import java.net.InetAddress;

import org.openmuc.jdlms.internal.connectionsettings.HdlcSettings;
import org.openmuc.jdlms.internal.connectionsettings.TcpSettings;
import org.openmuc.jdlms.internal.sessionlayer.SessionLayer;
import org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcAddress;
import org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcAddressPair;
import org.openmuc.jdlms.internal.sessionlayer.hdlc.HdlcLayer;
import org.openmuc.jdlms.internal.sessionlayer.wrapper.WrapperLayer;

public class TcpConnectionBuilder extends ConnectionBuilder<TcpConnectionBuilder> {

    private static final int DEFAULT_DLMS_PORT = 4059;
    private InetAddress inetAddress;
    private int tcpPort;
    private SessionLayerType sessionLayerType;

    /**
     * Construct a {@link TcpConnectionBuilder} with client ID 1, logical device address 16 and a default TCP port 4059.
     * 
     * @param inetAddress
     *            the Internet address of the remote meter.
     */
    public TcpConnectionBuilder(InetAddress inetAddress) {
        this.inetAddress = inetAddress;
        this.tcpPort = DEFAULT_DLMS_PORT;
        this.sessionLayerType = SessionLayerType.WRAPPER;
    }

    public TcpConnectionBuilder setInetAddress(InetAddress inetAddress) {
        this.inetAddress = inetAddress;
        return this;
    }

    public TcpConnectionBuilder setTcpPort(int tcpPort) {
        this.tcpPort = tcpPort;
        return this;
    }

    public TcpConnectionBuilder useHldc() {
        this.sessionLayerType = SessionLayerType.HDLC;
        return this;
    }

    public TcpConnectionBuilder useWrapper() {
        this.sessionLayerType = SessionLayerType.WRAPPER;
        return this;
    }

    @Override
    public DlmsConnection buildLnConnection() throws IOException {
        TcpSettingsImpl settings = new TcpSettingsImpl(this);

        SessionLayer transportLayer = buildSessionLayer(settings);

        DlmsLnConnection connection = new DlmsLnConnection(settings, transportLayer);

        connection.connect();

        return connection;
    }

    private SessionLayer buildSessionLayer(TcpSettingsImpl settings) throws IOException {
        switch (sessionLayerType) {
        case HDLC:
            return new HdlcLayer(settings);

        default:
        case WRAPPER:
            return new WrapperLayer(settings);
        }
    }

    @Override
    public DlmsConnection buildSnConnection() throws IOException {
        TcpSettingsImpl settings = new TcpSettingsImpl(this);
        SessionLayer transportLayer = buildSessionLayer(settings);

        DlmsSnConnection clientConnection = new DlmsSnConnection(settings, transportLayer);

        clientConnection.connect();

        return clientConnection;
    }

    public class TcpSettingsImpl extends SettingsImpl implements HdlcSettings, TcpSettings {

        private final InetAddress inetAddress;
        private final int tcpPort;
        private final HdlcAddressPair addressPair;

        public TcpSettingsImpl(TcpConnectionBuilder connectionBuilder) {
            super(connectionBuilder);
            this.inetAddress = connectionBuilder.inetAddress;
            this.tcpPort = connectionBuilder.tcpPort;

            HdlcAddress source = new HdlcAddress(clientId());
            HdlcAddress destination = new HdlcAddress(logicalDeviceId(), physicalDeviceId());
            this.addressPair = new HdlcAddressPair(source, destination);
        }

        @Override
        public InetAddress inetAddress() {
            return this.inetAddress;
        }

        @Override
        public int tcpPort() {
            return this.tcpPort;
        }

        @Override
        public HdlcAddressPair addressPair() {
            return this.addressPair;
        }

    }

    private enum SessionLayerType {
        HDLC,
        WRAPPER
    }
}
