/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.app;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.concurrent.TimeoutException;

import org.openmuc.jdlms.AccessResultCode;
import org.openmuc.jdlms.GetResult;
import org.openmuc.jdlms.datatypes.DataObject;

abstract class ConsoleApp {

    private static final String DATA_INPUT_FORMAT = "<Data_Type>:<Data>";

    private static final String POSSIBLE_DATA_TYPES = "(b)oolean / (f)loat / (d)ouble / (l)ong / (i)nteger / (o)ctet";

    private final BufferedReader inputReader;

    public ConsoleApp() {
        this.inputReader = new BufferedReader(new InputStreamReader(System.in));
    }

    public final void processRead() throws IOException {
        System.out.println("Enter: " + nameFormat());
        String requestParameter = inputReader.readLine();

        GetResult result;
        try {
            result = callGet(requestParameter);
        } catch (TimeoutException e) {
            e.printStackTrace();
            System.err.println("Failed to process read.");

            return;
        } catch (IllegalArgumentException e) {
            System.err.printf(e.getMessage());
            return;
        }
        AccessResultCode resultCode = result.getResultCode();

        if (resultCode == AccessResultCode.SUCCESS) {
            System.out.println("Result Code: " + result.getResultCode());

            DataObject resultData = result.getResultData();
            System.out.println(resultData.toString());
        }
        else {
            System.err.printf("Failed to read. AccessResultCode: %s%n", resultCode);
        }
    }

    public abstract void processScan() throws IOException;

    public final void processWrite() throws IOException {
        System.out.println("Enter: " + nameFormat());
        String address = inputReader.readLine();

        System.out.println("Enter: " + DATA_INPUT_FORMAT);
        System.out.println("possible data types: " + POSSIBLE_DATA_TYPES);

        String inputData;

        do {
            inputData = inputReader.readLine();
        } while (inputData == null);

        DataObject dataToWrite = buildDataObject(inputData);

        AccessResultCode resultCode = callSet(address, dataToWrite);
        if (resultCode == AccessResultCode.SUCCESS) {
            System.out.println("Result Code: " + resultCode);
        }
        else {
            System.err.printf("Failed to write. AccessResultCode: %s%n", resultCode);
        }
    }

    public void close() {
        try {
            this.inputReader.close();
        } catch (IOException e) {
            // ignore this here.
        }
    }

    protected abstract String nameFormat();

    protected abstract GetResult callGet(String requestParameter) throws IOException, TimeoutException;

    protected abstract AccessResultCode callSet(String requestParameter, DataObject dataToWrite) throws IOException;

    private DataObject buildDataObject(String line) {

        String[] arguments = line.split(":");

        if (arguments.length != 2) {
            throw new IllegalArgumentException(String.format("Wrong number of arguments. %s", DATA_INPUT_FORMAT));
        }

        String dataTypeString = arguments[0];
        String dataString = arguments[1];

        char datatype = dataTypeString.toUpperCase().charAt(0);

        DataObject dataObject;

        switch (datatype) {
        case 'S':
            short sData = Short.parseShort(dataString);
            dataObject = DataObject.newInteger16Data(sData);
            break;
        case 'I':
            int iData = Integer.parseInt(dataString);
            dataObject = DataObject.newInteger32Data(iData);
            break;
        case 'L':
            Long lData = Long.parseLong(dataString);
            dataObject = DataObject.newInteger64Data(lData);
            break;
        case 'F':
            float fData = Float.parseFloat(dataString);
            dataObject = DataObject.newFloat32Data(fData);
            break;
        case 'D':
            double dData = Double.parseDouble(dataString);
            dataObject = DataObject.newFloat64Data(dData);
            break;
        case 'B':
            boolean bData = Boolean.parseBoolean(dataString);
            dataObject = DataObject.newBoolData(bData);
            break;
        default:
            throw new IllegalArgumentException(String.format("Wrong data type. %s", POSSIBLE_DATA_TYPES));
        }

        return dataObject;
    }
}
