/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms;

import java.io.IOException;
import java.nio.charset.StandardCharsets;

import org.openmuc.jdlms.internal.ConfirmedMode;
import org.openmuc.jdlms.internal.connectionsettings.Settings;

/**
 * Class to create a {@link DlmsConnection}.
 * 
 * @param <T>
 *            the concrete connection builder.
 * 
 * @see TcpConnectionBuilder
 * @see SerialConnectionBuilder
 */
public abstract class ConnectionBuilder<T extends ConnectionBuilder<T>> {

    private final ConfirmedMode confirmedMode;

    private byte[] systemTitle;

    private int challengeLength;

    private long deviceId;

    private int responseTimeout;
    private final int messageFragmentTimeout;

    private String manufactureId;

    private int logicalDeviceId;

    private int clientId;

    private Authentication authentication;

    private int physicalDeviceId;

    /**
     * Create a new connection builder, with default settings.
     */
    public ConnectionBuilder() {
        this.confirmedMode = ConfirmedMode.CONFIRMED;

        this.systemTitle = new byte[] { 0x4d, 0x4d, 0x4d, 0, 0, 0, 0, 1 };

        this.challengeLength = 64;

        this.deviceId = 1;

        this.responseTimeout = 20000;
        this.messageFragmentTimeout = 5000;
        this.manufactureId = "";

        this.clientId = 16;
        this.logicalDeviceId = 1;

        this.authentication = Authentication.newNoneAuthentication();

        this.physicalDeviceId = 0;

    }

    /**
     * Set {@link Authentication}.
     * 
     * @param authentication
     *            authentication object
     * @return T the ConnectionBuilder
     */
    public T setAuthentication(Authentication authentication) {
        this.authentication = authentication;

        return self();
    }

    /**
     * Set the client ID which should be used to connect to the server.
     * 
     * <p>
     * Default client ID is 16 (Public Client).
     * </p>
     * 
     * @param clientId
     *            the client id.
     * @return T the ConnectionBuilder
     */
    public T setClientId(int clientId) {
        this.clientId = clientId;

        return self();
    }

    /**
     * Set the logical device ID of the logical device in the physical server.
     * 
     * <p>
     * Default logical device ID is 1 (Management Logical Device).
     * </p>
     * 
     * @param logicalDeviceId
     *            the logical device ID.
     * @return T the ConnectionBuilder
     */
    public T setLogicalDeviceId(int logicalDeviceId) {
        this.logicalDeviceId = logicalDeviceId;

        return self();
    }

    /**
     * Change the used challenge length. Allowed is from 8 to 64. Default is 64. <br>
     * <br>
     * If a number less then 8 or greater then 64 is chosen a IllegalArgumentException will be thrown.
     * 
     * @param challengeLength
     *            challenge length
     * @return T the ConnectionBuilder
     */
    public T setChallengeLength(int challengeLength) {
        int minLength = 8;
        int maxLength = 64;

        if (challengeLength < minLength || challengeLength > maxLength) {
            throw new IllegalArgumentException("Chalenge length has to be between " + minLength + " and " + maxLength);
        }

        this.challengeLength = challengeLength;
        return self();
    }

    /**
     * Set the used device ID.
     * 
     * @param deviceId
     *            the device ID
     * @return T the ConnectionBuilder
     * 
     */
    public T setDeviceId(long deviceId) {
        this.deviceId = deviceId;
        return self();
    }

    /**
     * Sets the physical device address.
     * 
     * <p>
     * This is only relevant if a connection over HDLC is used.
     * </p>
     * 
     * @param physicalDeviceAddress
     *            the physical device address.
     * 
     * @return T the ConnectionBuilder
     */
    public T setPhysicalDeviceAddress(int physicalDeviceAddress) {
        this.physicalDeviceId = physicalDeviceAddress;
        return self();
    }

    /**
     * Sets the time in milliseconds the client waits for a response.
     * 
     * @param responseTimeout
     *            time in milliseconds.
     * @return T the ConnectionBuilder
     */
    public T setResponseTimeout(int responseTimeout) {
        this.responseTimeout = responseTimeout;

        return self();
    }

    /**
     * Sets the manufacturer ID.
     * 
     * @param manufactureId
     *            the manufacturer ID
     * @return T the ConnectionBuilder
     */
    public T setManufactureId(String manufactureId) {
        this.manufactureId = manufactureId;

        return self();
    }

    /**
     * Set the client's system title. It consists of 8 bytes: 3 characters for the manufacturer ID and 5 bytes for the
     * device ID. The default is "MMM" (manufacturer ID) and 1 (device ID).
     * 
     * @param manufacturerId
     *            the manufacturer ID
     * @param deviceId
     *            the device ID
     * @return T the ConnectionBuilder
     */
    public T setSystemTitle(String manufacturerId, long deviceId) {
        this.deviceId = deviceId;

        this.systemTitle = new byte[8];

        byte[] manufacturerIdBytes = manufacturerId.getBytes(StandardCharsets.US_ASCII);
        systemTitle[0] = manufacturerIdBytes[0];
        systemTitle[1] = manufacturerIdBytes[1];
        systemTitle[2] = manufacturerIdBytes[2];
        systemTitle[3] = (byte) (deviceId >> 4 * 8);
        systemTitle[4] = (byte) (deviceId >> 3 * 8);
        systemTitle[5] = (byte) (deviceId >> 2 * 8);
        systemTitle[6] = (byte) (deviceId >> 1 * 8);
        systemTitle[7] = (byte) (deviceId);

        return self();
    }

    /**
     * Builds a new DLMS/COSEM LN Connection, which is connected to the remote meter.
     * 
     * @return A new LnClientConnection with the given settings.
     * @throws IOException
     *             if an error occurs, while connecting to the meter.
     */
    public abstract DlmsConnection buildLnConnection() throws IOException;

    /**
     * Builds a new DLMS/COSEM SN Connection, which is connected to the remote meter.
     * 
     * @return A new SnClientConnection with the given settings.
     * @throws IOException
     *             if an error occurs, while connecting to the meter.
     */
    public abstract DlmsConnection buildSnConnection() throws IOException;

    /**
     * Access the casted this reference.
     * 
     * @return the reference of the object.
     */
    @SuppressWarnings("unchecked")
    protected T self() {
        return (T) this;
    }

    protected abstract class SettingsImpl implements Settings {
        private final ConfirmedMode confirmedMode;
        private final byte[] systemTitle;
        private final int challengeLength;
        private final long deviceId;
        private final int responseTimeout;
        private final int messageFragmentTimeout;
        private final String manufactureId;
        private final int clientAccessPoint;
        private final int logicalDeviceAddress;
        private final Authentication authentication;
        private final int physicalDeviceId;

        public SettingsImpl(ConnectionBuilder<?> builder) {
            this.confirmedMode = builder.confirmedMode;
            this.systemTitle = builder.systemTitle;
            this.challengeLength = builder.challengeLength;
            this.deviceId = builder.deviceId;
            this.responseTimeout = builder.responseTimeout;
            this.messageFragmentTimeout = builder.messageFragmentTimeout;
            this.manufactureId = builder.manufactureId;
            this.clientAccessPoint = builder.clientId;
            this.logicalDeviceAddress = builder.logicalDeviceId;
            this.authentication = builder.authentication;
            this.physicalDeviceId = builder.physicalDeviceId;
        }

        @Override
        public Authentication authentication() {
            return this.authentication;
        }

        @Override
        public int challengeLength() {
            return this.challengeLength;
        }

        @Override
        public byte[] systemTitle() {
            return this.systemTitle;
        }

        @Override
        public long deviceId() {
            return this.deviceId;
        }

        @Override
        public int responseTimeout() {
            return this.responseTimeout;
        }

        @Override
        public int messageFragmentTimeout() {
            return this.messageFragmentTimeout;
        }

        @Override
        public String manufactureId() {
            return this.manufactureId;
        }

        @Override
        public ConfirmedMode confirmedMode() {
            return this.confirmedMode;
        }

        @Override
        public int logicalDeviceId() {
            return this.logicalDeviceAddress;
        }

        @Override
        public int physicalDeviceId() {
            return this.physicalDeviceId;
        }

        @Override
        public int clientId() {
            return this.clientAccessPoint;
        }

        @Override
        public void authenticationKey(byte[] authenticationKey) {
            this.authentication.authenticationKey = authenticationKey;
        }

        @Override
        public void globalEncryptionKey(byte[] globalEncryptionKey) {
            this.authentication.globalEncryptionKey = globalEncryptionKey;
        }

    }
}
