package org.openmuc.jdlms.internal;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collections;
import java.util.Set;

import org.openmuc.jdlms.AccessResultCode;
import org.openmuc.jdlms.IllegalAttributeAccessException;
import org.openmuc.jdlms.SelectiveAccessDescription;
import org.openmuc.jdlms.datatypes.DataObject;

public interface AttributeAccessor {

    DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
            throws IllegalAttributeAccessException;

    void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
            Long connectionId) throws IllegalAttributeAccessException;

    static class LogicalNameFakeAccessor implements AttributeAccessor {

        private final DataObject logicalName;

        public LogicalNameFakeAccessor(byte[] logicalName) {
            this.logicalName = DataObject.newOctetStringData(logicalName);
        }

        @Override
        public DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            return this.logicalName;
        }

        @Override
        public void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            throw new IllegalAttributeAccessException(AccessResultCode.READ_WRITE_DENIED);
        }

    }

    static class FieldAccessor implements AttributeAccessor {

        private final Field field;

        public FieldAccessor(Field field) {
            this.field = field;
        }

        @Override
        public DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            try {
                Object object = this.field.get(instance);

                if (object == null) {
                    return DataObject.newNullData();
                }

                return (DataObject) object;
            } catch (IllegalAccessException e) {
                throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
            }
        }

        @Override
        public void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            try {
                this.field.set(instance, newVal);
            } catch (IllegalAccessException e) {
                throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
            }
        }
    }

    static class MethodSetFieldGetAccessor implements AttributeAccessor {

        private final FieldAccessor fieldAccessor;
        private final MethodAttributeAccessor methodAccessor;

        public MethodSetFieldGetAccessor(Field field, Method setMethod) {
            this.fieldAccessor = new FieldAccessor(field);
            this.methodAccessor = new MethodAttributeAccessor(null, setMethod);
        }

        @Override
        public DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            return this.fieldAccessor.get(instance, selectiveAccessDescription, connectionId);
        }

        @Override
        public void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            this.methodAccessor.set(newVal, instance, selectiveAccessDescription, connectionId);
        }

    }

    static class FieldSetMethodGetAccessor implements AttributeAccessor {

        private final FieldAccessor fieldAccessor;
        private final MethodAttributeAccessor methodAccessor;

        public FieldSetMethodGetAccessor(Field field, Method getMethod) {
            this.fieldAccessor = new FieldAccessor(field);
            this.methodAccessor = new MethodAttributeAccessor(getMethod, null);
        }

        @Override
        public DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            return this.methodAccessor.get(instance, selectiveAccessDescription, connectionId);
        }

        @Override
        public void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            this.fieldAccessor.set(newVal, instance, selectiveAccessDescription, connectionId);
        }

    }

    public static class MethodAttributeAccessor implements AttributeAccessor {

        private final Method getMethod;
        private final Method setMethod;
        private final Set<Integer> accessSelectors;
        private final boolean containsGetId;
        private final boolean containsSetId;

        @SuppressWarnings("unchecked")
        public MethodAttributeAccessor(Method getMethod, Method setMethod) {
            this(getMethod, setMethod, Collections.EMPTY_SET);
        }

        public MethodAttributeAccessor(Method getMethod, Method setMethod, Set<Integer> accessSelectors) {
            this.getMethod = getMethod;
            this.setMethod = setMethod;
            this.accessSelectors = accessSelectors;

            this.containsGetId = methodHasConnectionIdParam(getMethod);
            this.containsSetId = methodHasConnectionIdParam(setMethod);
        }

        private boolean methodHasConnectionIdParam(Method method) {
            if (method == null) {
                return false;
            }

            Class<?>[] parameterTypes = method.getParameterTypes();

            if (parameterTypes.length == 0) {
                return false;
            }

            int lastIndex = parameterTypes.length - 1;

            return Long.class.isAssignableFrom(parameterTypes[lastIndex]);
        }

        @Override
        public DataObject get(Object instance, SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            Object[] parameter = buildGetParameter(selectiveAccessDescription, connectionId);
            try {
                Object object = this.getMethod.invoke(instance, parameter);

                if (object == null) {
                    return DataObject.newNullData();
                }

                return (DataObject) object;
            } catch (IllegalAccessException e) {
                throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
            } catch (InvocationTargetException e) {
                throw convert(e);
            }
        }

        private Object[] buildGetParameter(SelectiveAccessDescription selectiveAccessDescription, Long connectionId)
                throws IllegalAttributeAccessException {
            Object[] parameter = new Object[this.getMethod.getParameterTypes().length];
            setSelectiveAcccessDescription(selectiveAccessDescription, parameter, 0);

            if (this.containsGetId) {
                setConnectionId(connectionId, parameter);
            }

            return parameter;
        }

        private void setConnectionId(Long connectionId, Object[] parameter) {
            parameter[parameter.length - 1] = connectionId;
        }

        @Override
        public void set(DataObject newVal, Object instance, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            Object[] parameter = buildSetParameter(newVal, selectiveAccessDescription, connectionId);

            try {
                this.setMethod.invoke(instance, parameter);
            } catch (IllegalAccessException e) {
                throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
            } catch (InvocationTargetException e) {
                throw convert(e);
            }
        }

        private Object[] buildSetParameter(DataObject newVal, SelectiveAccessDescription selectiveAccessDescription,
                Long connectionId) throws IllegalAttributeAccessException {
            Object[] parameter = new Object[this.setMethod.getParameterTypes().length];

            parameter[0] = newVal;
            setSelectiveAcccessDescription(selectiveAccessDescription, parameter, 1);
            if (this.containsSetId) {
                setConnectionId(connectionId, parameter);
            }
            return parameter;
        }

        private void setSelectiveAcccessDescription(SelectiveAccessDescription selectiveAccessDescription,
                Object[] parameter, int indexOfAccess) throws IllegalAttributeAccessException {

            if (selectiveAccessDescription != null) {
                if (this.accessSelectors.isEmpty()) {
                    // TODO set correct reason.
                    throw new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
                }

                parameter[indexOfAccess] = selectiveAccessDescription;
            }
        }

        private IllegalAttributeAccessException convert(InvocationTargetException e)
                throws IllegalAttributeAccessException {
            Throwable targetException = e.getTargetException();
            if (!(targetException instanceof IllegalAttributeAccessException)) {
                return new IllegalAttributeAccessException(AccessResultCode.OTHER_REASON);
            }
            return (IllegalAttributeAccessException) targetException;
        }
    }

}
